# import os
# from rupo.morph.lstm import LSTMMorphoAnalysis
#
# dir_name = "/media/data/Datasets/Morpho/clean"
# lstm = LSTMMorphoAnalysis()
# lstm.prepare([os.path.join(dir_name, filename) for filename in os.listdir(dir_name)], "word_vocab.pickle", "gram.pickle")